interface EventListener {
  callback: Function;
  count: number;
}
type EventName = string;
type EventInfo = number;

export class EventCounter<T extends string> {
  events: Record<T, EventInfo> = {} as Record<T, EventInfo>;
  timer: Record<EventName, NodeJS.Timeout | undefined> = {};
  listeners: Record<EventName, EventListener[]> = {};

  constructor(public timeout: number = 200) {}

  private getMaxClicksLimit(eventName: T) {
    if(this.listeners[eventName] === undefined) return 0;
    return Math.max(...this.listeners[eventName].map((e) => e.count));
  }

  private emitListeners(eventName: T) {
    const counter = this.events[eventName];
    this.events[eventName] = 0;
    const listeners = this.listeners[eventName].filter(
      (e) => e.count === counter
    );
    for (const listener of listeners) {
      listener.callback();
    }
  }

  emit(eventName: T) {
    if (eventName in this.events === undefined) this.events[eventName] = 0;
    if (this.timer[eventName] !== undefined) {
      clearTimeout(this.timer[eventName]);
      this.timer[eventName] = undefined;
    }
    const timeoutCallback = () => {
      this.emitListeners(eventName);
      this.timer[eventName] = undefined;
    };
    this.events[eventName]++;
    console.log(this.events);
    if (this.events[eventName] >= this.getMaxClicksLimit(eventName))
      return timeoutCallback();

    this.timer[eventName] = setTimeout(timeoutCallback, this.timeout);
  }
  on(args: EventListener & { name: T }) {
    const { name: eventName, callback, count } = args;
    if (!(eventName in this.events)) this.events[eventName] = 0;
    if (!Object.prototype.hasOwnProperty.call(this.listeners, eventName))
      this.listeners[eventName] = [];
    this.listeners[eventName].push({ callback, count });
  }
}
